#include "machineReassignment.h"
#include <io.h>
#include <fcntl.h>

unsigned int MachineReassignment::readNumber(char *&pb){
  int c;
  unsigned int v = 0;
  c = *pb++;
  if (c > '0' && c <= '9'){
    v = c - '0';
  }
  while (' ' != (c=*pb++)){
    if (c == '\n' || c == '\t') break;
    v = (10 * v) + c - '0';
  }
  return v;
}


void MachineReassignment::readProblem(){
  ifstream ifs_assignment;
  int fid_model;
  unsigned int r, m, mm, s, d, p, b;
  unsigned int numDependencies(0);

  // read model
  fid_model = _open(modelFileName.c_str(), _O_RDONLY | _O_TEXT | _O_SEQUENTIAL);
  if (fid_model == -1){
    throw runtime_error("Failed to open the instance_file (" + modelFileName + ")");
  }
  int length = _filelength(fid_model);
  char *buffer = new char [length];
	_read(fid_model, buffer, length);
	_close(fid_model);
  char *pb = buffer;
  
  numResources = readNumber(pb);
  resources.resize(numResources);
  for (r = 0; r < numResources; ++r){
    resources[r].transient = (readNumber(pb) > 0);
    resources[r].weightLoadCost = readNumber(pb);
  }
  
  numMachines = readNumber(pb);
  machines.resize(numMachines);
  numNeighborhoods = 0;
  numLocations = 0;
  for (m = 0; m < numMachines; ++m){
    machines[m].neighborhood = readNumber(pb);
    machines[m].location = readNumber(pb);
    if (machines[m].location+1 > numLocations) numLocations = machines[m].location+1; 
    if (machines[m].neighborhood+1 > numNeighborhoods) numNeighborhoods = machines[m].neighborhood+1;
    machines[m].capacities.resize(numResources);
    for (r = 0; r < numResources; ++r){
      machines[m].capacities[r] = readNumber(pb);
    }
    machines[m].saftyCapacities.resize(numResources);
    for (r = 0; r < numResources; ++r){
      machines[m].saftyCapacities[r] = readNumber(pb);
    }
    machines[m].movingCosts.resize(numMachines);
    for (mm = 0; mm < numMachines; ++mm){
      machines[m].movingCosts[mm] = readNumber(pb);
    }
  }
  machinesInNeighborhoods.resize(numNeighborhoods);
  machinesInLocation.resize(numLocations);
  for (m = 0; m < numMachines; ++m){
    machinesInNeighborhoods[machines[m].neighborhood].push_back(m);
    machinesInLocation[machines[m].location].push_back(m);
  }

  numServices = readNumber(pb);
  services.resize(numServices);
  for (s = 0; s < numServices; ++s){
    services[s].id = s;
    services[s].spreadMin = readNumber(pb);
    services[s].numDependencies = readNumber(pb);
    numDependencies += services[s].numDependencies;
    services[s].dependencies.resize(services[s].numDependencies);
    for (d = 0; d < services[s].numDependencies; ++d){
      services[s].dependencies[d] = readNumber(pb);
    }
  }

  numProcesses = readNumber(pb);
  processes.resize(numProcesses);
  for (p = 0; p < numProcesses; ++p){
    processes[p].id = p;
    processes[p].service = readNumber(pb);
    processes[p].localId = services[processes[p].service].processes.size();
    services[processes[p].service].processes.push_back(p);
    processes[p].requirements.resize(numResources);
    for (r = 0; r < numResources; ++r){
      processes[p].requirements[r] = readNumber(pb);
    }
    processes[p].movingCost = readNumber(pb);
  }

  numBalances = readNumber(pb);
  balances.resize(numBalances);
  for (b = 0; b < numBalances; ++b){
    balances[b].resource1 = readNumber(pb);
    balances[b].resource2 = readNumber(pb);
    balances[b].target = readNumber(pb);
    balances[b].weight = readNumber(pb);
  }
  
  weightProcessMovingCost = readNumber(pb);
  weightServiceMovingCost = readNumber(pb);
  weightMachineMovingCost = readNumber(pb);

  delete [] buffer;
  
  // read original assignment
  ifs_assignment.open(assignmentFileName.c_str(), ios::in);
  if (ifs_assignment.fail()){
    throw runtime_error("Failed to open the original_solution_file (" + assignmentFileName + ")");
  }
  ifs_assignment.sync_with_stdio(false);

  for (p = 0; p < numProcesses; ++p){
    ifs_assignment >> processes[p].originalAssignment;
  }
  ifs_assignment.close();

  // print instance size
#if !SILENT
  cout << "# number of machines: " << numMachines << endl;
  cout << "# number of resources: " << numResources << endl;
  cout << "# number of processes: " << numProcesses << endl;
  cout << "# number of services: " << numServices  << endl;
  cout << "# number of neighborhoods: " << numNeighborhoods << endl;
  cout << "# number of dependencies: " << numDependencies << endl;
  cout << "# number of locations: " << numLocations << endl;
  cout << "# number of balance costs: " << numBalances << endl;
#endif
}

void MachineReassignment::outputSolution(){
#if !SILENT 
  cout.flush();
  cout << endl;
  cout.setf(ios_base::fixed);
  cout << "cpu time: " << cputime() << "(s)" << endl;
#endif
  cout << "penalty: " << solution.penalty.soft << endl;

  // output new assignment
  if (!invalid){
    ofstream sol;
    ostream *osp;
    sol.open(solutionFileName.c_str(), ios::out);
    if (sol.fail()){
      cout << "Failed to open the new_solution_file. Output to stdout instead." << endl;
      osp = &cout;
    }
    else osp = &sol;
    for (unsigned int p = 0; p < numProcesses; ++p){
      *osp << solution.assignment[p] << " ";
    }
    *osp << endl;
    sol.close();
  }
}
