#include "machineReassignment.h"

MachineReassignment::int64 MachineReassignment::ResourceConstraint::calculatePenalty(Solution &solution){
  // calculate penalty
  total = 0;
  for (unsigned int p = 0; p < con->solver->numProcesses; ++p){
    if (solution.assignment[p] == machine) total += requirement(p);
  }
  penalty = func(total - capacity);
  return penalty;
}

MachineReassignment::int64 MachineReassignment::ResourceConstraint::updatePenalty(Solution &solution, Solver::Move *move){
  ppenalty = penalty;
  switch(move->type){
  case Solver::Shift:
    {
      Solver::ShiftMove *shift = static_cast<Solver::ShiftMove*>(move);
      if (shift->machine == machine) total += requirement(shift->process);
      else{
        if (solution.assignment[shift->process] != machine){
          throw runtime_error("Error in ResourceConstraint:updatePenalty");
        }
        total -= requirement(shift->process);
      }
      break;
    }
  case Solver::Swap:
    {
      Solver::SwapMove *swap = static_cast<Solver::SwapMove*>(move);
      unsigned int machine1 = solution.assignment[swap->process1];
      unsigned int machine2 = solution.assignment[swap->process2];
      if (machine1 == machine && machine2 != machine) total += requirement(swap->process2)-requirement(swap->process1);
      if (machine1 != machine && machine2 == machine) total += requirement(swap->process1)-requirement(swap->process2);
      break;
    }
  }
  penalty = func(total - capacity);
  return penalty - ppenalty;
}

void MachineReassignment::ResourceConstraint::calculateDiff(Solution &solution){
  // calculate diff
  //shift
  for (unsigned int p = 0; p < con->solver->numProcesses; ++p){
    if (solution.assignment[p] == machine){
      setAdjustmentForProcess(p, func(total-requirement(p)-capacity)-penalty);
      setShiftDiffPenalty(p, machine, -(func(total-requirement(p)-capacity)-penalty));
    }
    else{
      setShiftDiffPenalty(p, machine, func(total+requirement(p)-capacity)-penalty);
    }
  }
  //swap
  for (unsigned int s = 0; s < con->solver->numServices; ++s){
    unsigned int localId = 0;
    for (vector<Solver::SwapMove>::iterator ite = con->solver->moves.swapMoves[s].begin(); ite != con->solver->moves.swapMoves[s].end(); ++ite, ++localId){
      if (solution.assignment[ite->process1] == machine){
        if (solution.assignment[ite->process2] != machine){
          setSwapDiffPenalty(s, localId, func(total+requirement(ite->process2)-requirement(ite->process1)-capacity) - penalty);
        }
      }
      else if (solution.assignment[ite->process2] == machine){
          setSwapDiffPenalty(s, localId, func(total+requirement(ite->process1)-requirement(ite->process2)-capacity) - penalty);
      }
    }
  }
}

void MachineReassignment::ResourceConstraint::prepare(Solver *solver){
  con->addConstraint(this);
  involve(inf, machine);
}

void MachineReassignment::ResourceConstraint::prepare2(Solver *solver){
  calculatePenalty(solver->current);
  capacity = capacity0 - total0 + total;
  totalBackup = total;
}
void MachineReassignment::ResourceConstraint::postprocess(Solution &solution){
  calculatePenalty(solution);
  total0 += total - totalBackup;
}


// ----- Capacity/Transient Usage Constraints -----
void MachineReassignment::Constraints::initCapacityConstraints(){
  capacityConstraints.resize(mr->numMachines);
  for (unsigned int m = 0; m < mr->numMachines; ++m){
    capacityConstraints[m].resize(mr->numResources);
    for (unsigned int r = 0; r < mr->numResources; ++r){
      ResourceConstraint *constraint = new CapacityConstraint(this, inf, m, r);
      capacityConstraints[m][r] = constraint;
      constraint->capacity0 = mr->machines[m].capacities[r];
      unsigned int adjust = 0;
      for (unsigned int p = 0; p < mr->numProcesses; ++p){
        if (mr->resources[r].transient && mr->processes[p].originalAssignment == m){
          adjust += mr->processes[p].requirements[r];
        }  
      }
      constraint->capacity0 -= adjust;
    }
  }
  for (unsigned int p = 0; p < mr->numProcesses; ++p){
    unsigned int m = mr->solution.assignment[p];
    for (unsigned int r = 0; r < mr->numResources; ++r){
      if (!mr->resources[r].transient || mr->processes[p].originalAssignment != m){
        capacityConstraints[m][r]->total0 += mr->processes[p].requirements[r];
      }
    }
  }
}
void MachineReassignment::Constraints::prepareCapacityConstraints(){
  for (unsigned int m = 0; m < mr->numMachines; ++m){
    for (unsigned int r = 0; r < mr->numResources; ++r){
      ResourceConstraint *constraint = capacityConstraints[m][r];
      constraint->prepare(solver);
      constraint->prepare2(solver);
    }
  }
}
void MachineReassignment::Constraints::postprocessCapacityConstraints(Solution &solution){
  for (unsigned int m = 0; m < mr->numMachines; ++m){
    for (unsigned int r = 0; r < mr->numResources; ++r){
      capacityConstraints[m][r]->postprocess(solution);
    }
  }
}

// ----- Load Cost -----
void MachineReassignment::Constraints::initLoadCosts(){
  loadCosts.resize(mr->numMachines);
  for (unsigned int m = 0; m < mr->numMachines; ++m){
    loadCosts[m].resize(mr->numResources);
    for (unsigned int r = 0; r < mr->numResources; ++r){
      loadCosts[m][r] = new LoadCost(this, mr->resources[r].weightLoadCost, m, r);
      loadCosts[m][r]->capacity0 = mr->machines[m].saftyCapacities[r];
    }
  }
  for (unsigned int p = 0; p < mr->numProcesses; ++p){
    unsigned int m = mr->solution.assignment[p];
    for (unsigned int r = 0; r < mr->numResources; ++r){
      loadCosts[m][r]->total0 += mr->processes[p].requirements[r];
    }
  }
}

void MachineReassignment::Constraints::prepareLoadCosts(){
  for (unsigned int m = 0; m < mr->numMachines; ++m){
    for (unsigned int r = 0; r < mr->numResources; ++r){
      ResourceConstraint *constraint = loadCosts[m][r];
      constraint->prepare(solver);
      constraint->prepare2(solver);
    }
  }
}
void MachineReassignment::Constraints::postprocessLoadCosts(Solution &solution){
  for (unsigned int m = 0; m < mr->numMachines; ++m){
    for (unsigned int r = 0; r < mr->numResources; ++r){
      loadCosts[m][r]->postprocess(solution);
    }
  }
}

// ----- Balance Cost -----
void MachineReassignment::Constraints::initBalanceCosts(){
  balanceCosts.resize(mr->numBalances);
  for (unsigned int b = 0; b < mr->numBalances; ++b){
    balanceCosts[b].resize(mr->numMachines);
    unsigned int r = mr->balances[b].resource1;
    unsigned int r2 = mr->balances[b].resource2;
    for (unsigned int m = 0; m < mr->numMachines; ++m){
      balanceCosts[b][m] = new BalanceCost(this, mr->balances[b].weight, m, b, r, r2);
      balanceCosts[b][m]->capacity0 = 
        -1 * static_cast<int64>(mr->balances[b].target)*
        mr->machines[m].capacities[r] + mr->machines[m].capacities[r2];
    }
  }
  for (unsigned int p = 0; p < mr->numProcesses; ++p){
    unsigned int m = mr->solution.assignment[p];
    for (unsigned int b = 0; b < mr->numBalances; ++b){
      unsigned int r = mr->balances[b].resource1;
      unsigned int r2 = mr->balances[b].resource2;
      balanceCosts[b][m]->total0 +=
        -1 * static_cast<int64>(mr->balances[b].target)*mr->processes[p].requirements[r] 
        + mr->processes[p].requirements[r2];
    }
  }
}
void MachineReassignment::Constraints::prepareBalanceCosts(){
  for (unsigned int b = 0; b < mr->numBalances; ++b){
    unsigned int r = mr->balances[b].resource1;
    unsigned int r2 = mr->balances[b].resource2;
    for (unsigned int m = 0; m < mr->numMachines; ++m){
      ResourceConstraint *constraint = balanceCosts[b][m];
      constraint->prepare(solver);
      constraint->prepare2(solver);
    }
  }
}
void MachineReassignment::Constraints::postprocessBalanceCosts(Solution &solution){
  for (unsigned int b = 0; b < mr->numBalances; ++b){
    for (unsigned int m = 0; m < mr->numMachines; ++m){
      balanceCosts[b][m]->postprocess(solution);
    }
  }
}
