#include "machineReassignment.h"

void MachineReassignment::ServiceMovingCost::makeListsAndTree(){
  for (unsigned int k = 0; k <= ub; ++k){
    services[k].clear();
  }
  tree.clear();
  for (unsigned int s = 0; s < con->mr->numServices; ++s){ 
    unsigned int k = counts[s];
    sindex[s] = services[k].size();
    services[k].push_back(s);
    if (services[k].size() == 1){
      cindex[k] = tree.insert(k).first;
    }
  }
  penalty = (*tree.begin());
}

MachineReassignment::int64 MachineReassignment::ServiceMovingCost::calculatePenalty(Solution &solution){
  for (unsigned int s = 0; s < con->mr->numServices; ++s){
    counts[s] = 0;
    if (con->mr->services[s].group == con->group){
      vector<unsigned int> &processes = con->solver->services[con->mr->services[s].id2].processes;
      for (vector<unsigned int>::iterator ite = processes.begin(); ite != processes.end(); ++ite){
        if (solution.assignment[*ite] != con->solver->processes[*ite].originalAssignment) ++counts[s];
      }
    }
    else{
    vector<unsigned int> &processes = con->mr->services[s].processes;
      for (vector<unsigned int>::iterator ite = processes.begin(); ite != processes.end(); ++ite){
        if (con->mr->solution.assignment[*ite] != con->mr->processes[*ite].originalAssignment) ++counts[s];
      }
    }
  }
  makeListsAndTree();
  return penalty;
}

void MachineReassignment::ServiceMovingCost::changeCount(unsigned int service, int diff){
  if (diff == 0) return;
  unsigned int count = counts[service];
  if (services[count].size() > 1){
    unsigned int tail = *services[count].rbegin();
    services[count][sindex[service]] = tail;
    services[count].pop_back();
    sindex[tail] = sindex[service];
  }
  else{ // count == 1
    services[count].clear();
    tree.erase(cindex[count]);
  }
  counts[service] += diff;
  count = counts[service];
  sindex[service] = services[count].size();
  services[count].push_back(service);
  if (services[count].size() == 1){
    cindex[count] = tree.insert(count).first;
  }
}


int MachineReassignment::ServiceMovingCost::swapDiff(Solution &solution, Solver::SwapMove *swap){
  unsigned int machine1 = solution.assignment[swap->process1];
  unsigned int machine2 = solution.assignment[swap->process2];
  unsigned int original1 = con->solver->processes[swap->process1].originalAssignment;
  unsigned int original2 = con->solver->processes[swap->process2].originalAssignment;
  int diff = 0;
  if (machine1 == original1 && machine2 != original1) ++diff;
  if (machine1 != original1 && machine2 == original1) --diff;
  if (machine2 == original2 && machine1 != original2) ++diff;
  if (machine2 != original2 && machine1 == original2) --diff;
  return diff;
}

  MachineReassignment::int64 MachineReassignment::ServiceMovingCost::updatePenalty(Solution &solution, Solver::Move *move){
  ppenalty = penalty;
  switch(move->type){
  case Solver::Shift:
    {
      Solver::ShiftMove *shift = static_cast<Solver::ShiftMove*>(move);
      unsigned int service = con->solver->services[con->solver->processes[shift->process].service].id;
      unsigned int original = con->solver->processes[shift->process].originalAssignment;
      if (shift->machine == original){
        changeCount(service, -1);
      }
      else if (solution.assignment[shift->process] == original){
        changeCount(service, 1);
      }
      break;
    }
  case Solver::Swap:
    {
      Solver::SwapMove *swap = static_cast<Solver::SwapMove*>(move);
      int diff = swapDiff(solution, swap);
      unsigned int service = con->solver->services[con->solver->processes[swap->process1].service].id;
      changeCount(service, diff);
      break;
    }
  }
  penalty = *tree.begin();
  return penalty - ppenalty;
}

void MachineReassignment::ServiceMovingCost::calculateDiff(Solution &solution){
  // calculate diff
  // shift
  vector<unsigned int>::iterator site, pite;
  unsigned int maxCount = static_cast<unsigned int>(penalty);
  for (site = services[maxCount].begin(); site != services[maxCount].end(); ++site){
    if (con->mr->services[*site].group != con->group) continue;
    unsigned int s = con->mr->services[*site].id2;
    for (pite = con->solver->services[s].processes.begin(); pite != con->solver->services[s].processes.end(); ++pite){
      if (solution.assignment[*pite] != con->solver->processes[*pite].originalAssignment){
        if (services[maxCount].size() == 1) setShiftDiffPenalty(*pite, con->solver->processes[*pite].originalAssignment, -1);
      }
      else{
        setAdjustmentForProcess(*pite, 1);
        setShiftDiffPenalty(*pite, con->solver->processes[*pite].originalAssignment, -1);
      }
    }
  }
  // swap
  unsigned int secondMaxCount(0);
  set<int64,greater<int64> >::iterator ite = tree.begin();
  if (++ite != tree.end()) secondMaxCount = static_cast<unsigned int>(*ite);
  for (site = services[maxCount].begin(); site != services[maxCount].end(); ++site){
    if (con->mr->services[*site].group != con->group) continue;
    unsigned int s = con->mr->services[*site].id2;
    for (vector<Solver::SwapMove>::iterator swap = con->solver->moves.swapMoves[s].begin(); swap != con->solver->moves.swapMoves[s].end(); ++swap){
      int diff = swapDiff(solution, &(*swap));
      if (diff > 0){
        setSwapDiffPenalty(&(*swap), diff);
      }
      else if (diff < 0 && services[maxCount].size() == 1){
        setSwapDiffPenalty(&(*swap), maxCount - max(secondMaxCount, maxCount+diff));
      }
    }
  }
}

void MachineReassignment::ServiceMovingCost::prepare(Solver *solver){
  con->addConstraint(this);
  for (unsigned int p = 0; p < solver->numProcesses; ++p){
    involve(p, inf);
  }
}

// ----- Service Moving Cost -----
void MachineReassignment::Constraints::initServiceMovingCost(){
  serviceMovingCost = new ServiceMovingCost(this, mr->weightServiceMovingCost);
}
void MachineReassignment::Constraints::prepareServiceMovingCost(){
  serviceMovingCost->prepare(solver);
}
